import { createStore, combineReducers } from 'redux';
import _ from 'lodash';
import { createDraft, finishDraft, enablePatches, applyPatches } from 'immer';

function ownKeys(object, enumerableOnly) {
  var keys = Object.keys(object);

  if (Object.getOwnPropertySymbols) {
    var symbols = Object.getOwnPropertySymbols(object);

    if (enumerableOnly) {
      symbols = symbols.filter(function (sym) {
        return Object.getOwnPropertyDescriptor(object, sym).enumerable;
      });
    }

    keys.push.apply(keys, symbols);
  }

  return keys;
}

function _objectSpread2(target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i] != null ? arguments[i] : {};

    if (i % 2) {
      ownKeys(Object(source), true).forEach(function (key) {
        _defineProperty(target, key, source[key]);
      });
    } else if (Object.getOwnPropertyDescriptors) {
      Object.defineProperties(target, Object.getOwnPropertyDescriptors(source));
    } else {
      ownKeys(Object(source)).forEach(function (key) {
        Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key));
      });
    }
  }

  return target;
}

function _typeof(obj) {
  "@babel/helpers - typeof";

  if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") {
    _typeof = function (obj) {
      return typeof obj;
    };
  } else {
    _typeof = function (obj) {
      return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj;
    };
  }

  return _typeof(obj);
}

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function _toConsumableArray(arr) {
  return _arrayWithoutHoles(arr) || _iterableToArray(arr) || _unsupportedIterableToArray(arr) || _nonIterableSpread();
}

function _arrayWithoutHoles(arr) {
  if (Array.isArray(arr)) return _arrayLikeToArray(arr);
}

function _iterableToArray(iter) {
  if (typeof Symbol !== "undefined" && iter[Symbol.iterator] != null || iter["@@iterator"] != null) return Array.from(iter);
}

function _unsupportedIterableToArray(o, minLen) {
  if (!o) return;
  if (typeof o === "string") return _arrayLikeToArray(o, minLen);
  var n = Object.prototype.toString.call(o).slice(8, -1);
  if (n === "Object" && o.constructor) n = o.constructor.name;
  if (n === "Map" || n === "Set") return Array.from(o);
  if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen);
}

function _arrayLikeToArray(arr, len) {
  if (len == null || len > arr.length) len = arr.length;

  for (var i = 0, arr2 = new Array(len); i < len; i++) arr2[i] = arr[i];

  return arr2;
}

function _nonIterableSpread() {
  throw new TypeError("Invalid attempt to spread non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.");
}

function getPatches(_ref) {
  var store = _ref.store,
      payload = _ref.payload,
      action = _ref.action,
      path = _ref.path;
  var state = store.getState(); // default state from both undefined and null to {} if we have a path

  if (path) {
    state = state === undefined || state === null ? {} : state;
    var notObject = _typeof(state) !== 'object';
    var notFunction = typeof state !== 'function';

    if (notObject && notFunction) {
      var error = "To use a path, your store must be either an Object, Array or Function, but got ".concat(_typeof(state));
      console.error(error, state);
      throw new Error(error);
    }
  } // if we don't throw here, the function continues "successfully" and throws obliquely later. :s


  if (action) {
    var _notObject = _typeof(action) !== 'object';

    var _notFunction = typeof action !== 'function';

    if (_notObject && _notFunction) {
      var _error = "Actions must be either an Object or Function, but got ".concat(_typeof(action));

      console.error(_error, action);
      throw new Error(_error);
    }
  } // Try to create a draft. If we can't, it's a value we can't proxy (i.e. null, boolean, number)


  var draft = safeCreateDraft(state); // this is a little bit of a mess.
  // We're trying to take four different shapes and convert them into one shape
  // `action`  (function) -> action() gives us the `returnAction`
  // `action`  (non-fn)   -> action is the `returnAction`
  // `payload` (function) -> payload() gives us the payload of `returnAction`
  // `payload` (non-fn)   -> payload is the payload of `returnAction`
  // if we were passed an action, process the payload
  // if we were passed a payload, process the payload and stuff it in an action

  var actionOrPayload = action || payload;
  var isFunction = typeof actionOrPayload === 'function';
  var returnAction;

  if (isFunction) {
    returnAction = actionOrPayload(path ? _.get(draft, path) : draft);
  } else {
    returnAction = actionOrPayload;
  }

  if (!action) {
    returnAction = {
      payload: returnAction
    };
  } // handle the payload so that we can do it all in one step


  if (path) {
    returnAction.payload = _.set(draft, path, returnAction.payload);
  }

  var patches = safeBuildPatches(returnAction.payload);
  if (!patches) _.set(returnAction, 'meta.nbpr', 'REPLACE');else {
    returnAction.payload = patches;

    _.set(returnAction, 'meta.nbpr', 'UPDATE');
  }
  return returnAction;
}

function safeCreateDraft(state) {
  var draft;

  try {
    draft = createDraft(state);
  } catch (error) {
    if (error.message !== '[Immer] First argument to `createDraft` must be a plain object, an array, or an immerable object') {
      throw error;
    }

    draft = state;
  }

  return draft;
}

function safeBuildPatches(draft) {
  var changes = [];

  try {
    finishDraft(draft, function (patches) {
      return changes.push.apply(changes, _toConsumableArray(patches));
    });
  } catch (error) {
    // or (if it wasn't a draft) replace it
    if (error.message !== "[Immer] First argument to `finishDraft` must be a draft returned by `createDraft`") {
      throw error;
    }

    return null;
  }

  return changes;
}

/**
 * 
 * Generates the action to fire.
 * 
 * @param {Object} firingParameters   An object that contains the data needed to fire an update action
 * @param {*} firingParameters.store   The redux store. 
 * @param {Array} [firingParameters.path]   A sub-path to pass to the valueFn or be set to during the reducer phase
 * @param {*|Function} firingParameters.payload   If a function, will be run during the reducer phase. If a value, will be set during the reducer phase.
 * @param {String|Object} [firingParameters.action]   A set of params from the user that can add/remove params in the action.
 */

function fireUpdateAction(_ref) {
  var store = _ref.store,
      path = _ref.path,
      payload = _ref.payload,
      action = _ref.action;
  // properties that are defaulted that can be overridden
  var canOverride = {
    type: undefined // 'SET_STORE' 

  }; // properties that are defaulted that can't be overridden

  var cantOverride = {
    payload: undefined,
    // what we're replacing
    meta: {// nbpr: 'UPDATE' || 'REPLACE' 
    }
  }; // turn path into an array

  var pathAsArray;

  if (typeof path === 'string' || Array.isArray(path)) {
    pathAsArray = _.toPath(path);
    if (!pathAsArray.length) pathAsArray = undefined;
  } // pre-run the function 


  var baseAction = action ? getPatches({
    store: store,
    action: action,
    path: path
  }) : getPatches({
    store: store,
    payload: payload,
    path: path
  });

  if (_.get(baseAction, 'meta.nbpr') === 'UPDATE') {
    canOverride.type = 'UPDATE';
  } // We do a replace in every other case. This is so that we can replace with falsy values.
  else {
    canOverride.type = 'REPLACE';
  } // customize the action
  // customize the action type, first


  canOverride.type += "_".concat((pathAsArray ? pathAsArray[0] : 'store').toUpperCase());

  var completeAction = _.merge(canOverride, baseAction, cantOverride);

  store.dispatch(completeAction);
  return pathAsArray ? store.get(pathAsArray) : store.get();
}

/**
 * 
 * A setState-style interface to a Redux store. If you need to customize the action fire, use .action
 * 
 * @param {String|Array} [path]   A lodash-style path value to set. Optional.
 * @param {*|Function} payload   Data you want to replace the store with, or a function that produces the changes you want to make to the store.
 */

function set(path, payload) {
  var store = this; // if we only have one argument, the signature is `.set(payload)`

  if (arguments.length === 1) {
    path = undefined;
    payload = arguments[0];
    return fireUpdateAction({
      store: store,
      path: path,
      payload: payload
    });
  } // if we have two arguments, the signature is `.set(path, payload)`
  else if (arguments.length === 2) {
    return fireUpdateAction({
      store: store,
      path: path,
      payload: payload
    });
  }

  if (process.env.NODE_ENV !== 'production') {
    throw new Error(".set expects one parameter ( .set(payload) ), two parameters ( .set(path, payload) ), but received ".concat(arguments.length, ": ").concat([].join.call(arguments, ', '), ")"));
  }
}

/**
 * 
 * Get the store or a part of the store.
 * 
 * @param {String|Array} [path]   A lodash-style path value to get. Optional.
 * @param {*} [defaultValue]   A value to return as the default value if no value exists on the path. Optional. Path is required if this value is set.
 */

function get(path, defaultValue) {
  var store = this; // .get()

  if (arguments.length === 0) {
    return store.getState();
  } // .get(path) or .get(path, defaultValue)
  else if (arguments.length === 1 || arguments.length === 2) {
    return _.get(store.getState(), path, defaultValue);
  }

  if (process.env.NODE_ENV !== 'production') {
    throw new Error(".get expects zero parameters ( .get() ), one parameter ( .get(path) ), or two parameters ( .get(path, defaultValue), but received ".concat(arguments.length, ": ").concat([].join.call(arguments, ', '), ")"));
  }
}

/**
 * 
 * A setState-style interface to a Redux store with the ability to customize the action fired.
 * 
 * @param {String|Array} [path]   A lodash-style path value to set. Optional.
 * @param {Object|Function} action   An object representing the action you want to fire, or a function that produces the action you want to fire.
 * @param {Object} action   Any properties to add to the actions to be fired to make these changes.
 * 
 */

function action(path, action) {
  var store = this; // if we only have two arguments, the signature is `.set(action, action)`

  if (arguments.length === 1) {
    path = undefined;
    action = arguments[0];
    return fireUpdateAction({
      store: store,
      path: path,
      action: action
    });
  } else if (arguments.length === 2) {
    return fireUpdateAction({
      store: store,
      path: path,
      action: action
    });
  }

  if (process.env.NODE_ENV !== 'production') {
    throw new Error(".action expects one parameter ( .action(action) ), or two parameters ( .action(path, action) ), but received ".concat(arguments.length, ": ").concat([].join.call(arguments, ', '), ")"));
  }
}

enablePatches();
function nbprReducer() {
  var state = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

  var _ref = arguments.length > 1 ? arguments[1] : undefined,
      meta = _ref.meta,
      payload = _ref.payload;

  var returnState = state;

  if (meta.nbpr === 'UPDATE') {
    if (state === undefined || state === null) {
      returnState = {};
    }

    returnState = applyPatches(returnState, payload);
  } else if (meta.nbpr === 'REPLACE') {
    returnState = payload;
  }

  return returnState;
}

var stores = {};
/**
 * 
 * A createStore replacement. Does all the same things as createStore 
 * (and accepts very similar arguments), and initializes no boilerplate redux.
 * 
 * @param {Object} config   A config object used to reference, create and recreate the store.
 * @param {String} [key='global']   A key used to save and get the store.
 * @param {Object|Function} [config.reducer]   A reducer function or an object of reducers function to include alongside no-boilerplate-redux
 * @param {Object} [config.reducerCombiner=combineReducers]   If reducer is an object and not a function, the function we use to combine them. Defaults to redux's combineReducers
 * @param {Object} [config.preloadedState]   An object of default state; does not require `reducer` param to function
 * @param {Function} [config.enhancer]   A function to enhance your store with third-party capabilities. (see  https://github.com/reduxjs/redux/blob/master/docs/api/createStore.md)
 */

var makeStore = function makeStore() {
  var _ref = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {},
      _ref$key = _ref.key,
      key = _ref$key === void 0 ? 'global' : _ref$key,
      userReducer = _ref.reducer,
      preloadedState = _ref.preloadedState,
      enhancer = _ref.enhancer;

  if (stores[key]) {
    throw new Error("Store with key \"".concat(key, "\" already initialized! If you wanted a second store please use another key, and if you wanted to clear the store, use `store().set()`."));
  }

  var passThroughArguments = [preloadedState, enhancer];

  function reducerRouter(state, action) {
    var nbprMeta = _.get(action, 'meta.nbpr', Symbol["for"]('nbpr-meta-not-set')); // if meta.nbpr is not set, don't bother with running the nbprReducer


    if (nbprMeta === Symbol["for"]('nbpr-meta-not-set')) return userReducer ? userReducer(state, action) : state; // if the user didn't provide a reducer, just to nbpr

    if (!userReducer) return nbprReducer(state, action); // if the user provided one, run it to make sure we don't remove anything we're not supposed to.

    return userReducer(nbprReducer(state, action), _objectSpread2(_objectSpread2({}, action), {}, {
      type: '' // clear the type so we don't accidentally trigger some default reducers

    }));
  }

  var store = createStore.apply(void 0, [reducerRouter].concat(passThroughArguments));
  store.set = set.bind(store);
  store.get = get.bind(store);
  store.action = action.bind(store);
  stores[key] = store;
  return store;
};
/**
 * 
 * Get the store referenced by the provided key.
 * 
 * @param {String} [storeKey='global']  The key of the store. If not provided, use the global store. 
 */

function getStore() {
  var storeKey = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : 'global';
  return stores[storeKey];
}

function nullReducer() {
  var state = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : null;
  return state;
}
/**
 * 
 * @param {Object} reducers   An object whose values correspond to different reducer 
 *    functions that need to be combined into one. One handy way to obtain it is to 
 *    use ES6 `import * as reducers` syntax. The reducers may never return undefined 
 *    for any action. Instead, they should return their initial state if the state 
 *    passed to them was undefined, and the current state for any unrecognized action.
 * @param {Function} [combiner]   A function used to combine the reducers object into
 *    a reducer function. Defaults to redux's combineReducers function.
 */


function makeReducer(reducers) {
  var combiner = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : combineReducers;
  var reducerKeys = Object.keys(reducers).sort().join();
  var reducer = combiner(reducers);
  return function nbprCombination(state, action) {
    var newReducerKeys = Object.keys(state).sort();

    if (reducerKeys !== newReducerKeys.join()) {
      newReducerKeys.forEach(function (key) {
        if (!reducers[key]) reducers[key] = nullReducer;
      });
      reducerKeys = newReducerKeys.join();
      reducer = combiner(reducers);
    }

    return reducer(state, action);
  };
}

export { makeReducer, makeStore, getStore as store };
